#!php
<?php
/*
#
# Auteur: Achraf cherti (alias Asher256)
# Email:  achrafcherti@gmail.com
# 
# Dernire version: http://asher256.tuxfamily.org/cmak
#
#Licence :
#==========
# Ce code source est distribu sous licence GNU General Public
# Licence v2. Vous pouvez de le distribuer, l'utiliser, l'tudier
# ou l'amliorer. Pour plus d'informations lisez le fichier
# COPYING.txt ou COPYING-FR.txt distribus avec le code source.
#
# Description :
#==============
# CMak vous permet de crer un Makefile rapidement pour votre
# projet ou pour un besoin de compilation rapide (Si vous 
# trouvez sur internet par exemple un projet c/c++ qui ne
# contient aucun Makefile).
#
# La fonction "detect lib":
#==========================
# Si l'option --detect-lib est dclare, le programme 
# dtecte les bibliothques de sous programmes requises
# par le code source en utilisant les directives
# "#include" dclares dans les fichiers c/c++.
#
# Exemple: S'il y a un include "math.h", "-lm" sera 
#          automatiquement ajout aux LDFLAGS.
#
# Le fichier cmak.cfg :
#======================
#
# Ce fichier est utilis pour permettre de trouver la
# correspondance entre header<->cflags/ldflags (information
# utilise par --detect-lib). Ce fichier peut tre mis dans 
# plusieurs endroits:
#
#   * Sous Unix (par ex. Linux ou MacOSX):
#                ./cmak.cfg
#                $HOME/.cmak.cfg
#                /etc/cmak.cfg
#                /usr/share/cmak/cmak.cfg
#
#  * Sous Windows:
#		.\cmak.cfg
#		c:\cmak.cfg
#
# NB: La lecture des fichiers cmak.cfg se fait dans l'ordre
#     inverse de la liste ci-dessus. Cela veut dire que sous UNIX,
#     /usr/share/cmak/cmak.cfg est le premier  tre lu (s'il existe).
#     Mais attention, c'est le dernier lu qui est le premier servi. Ainsi,
#     les correspondances entre Header<->CFLAGS/LDFLAGS de "./cmak.cfg" sont
#     prioritaires sur celles de "$HOME/.cmak.cfg" et celle de "/etc/cmak.cfg" 
#     sont prioritaires sur celles de "/usr/share/cmak/cmak.cfg".
#     
# NB: En vrit, le premier cmak.cfg lu c'est celui qui est 
#     dclar avec l'option "--cfg". J'ai prfr le dire dans
#     le deuxime note pour plus de claret dans l'explication :-)
#

use strict;     # Imposer la dclaration des variables
use warnings;   # Activer les warnings

use File::Basename; # dirname() basename()
*/

# Const
$VERSION="0.9.6 (4/2006)";         # La version du programme
$valid_extention=array("c","cpp","cxx","cc"); # Les extentions valides
$cc="gcc:g++"; #compilateur c et c++

# Les variables modifiables par les arguments (et leurs valeurs par dfaut)
$default_path="";
$makefile="Makefile";
$detect_lib=0;
$verbose=0;
$interactive=0;
$executable="";
$ldflags="";
$cflags="";
$objdir="";
$cmak_cfg="";
$cpp_file_list=array(); # liste de fichiers C/CPP passs dans les arguments

# Les variables modifies pour une utilisation interne
$include_path_list = array(); # Liste des chemins pointant vers les includes
$cpp_source=0; # =1 si un fichier de la liste est c++

# Correspondance entre Headers/CFLAGS/LDFLAGS (l'option --detect-lib)
$cmak_header_list = array();
$cmak_cflags_list = array();
$cmak_ldflags_list = array();

$define=array("gcc", "unix"); # Toujours dclare (car il n'y a que gcc qui est support par cmak pour le moment... TODO)


/* rcuprer le systeme detect 
# Defines internes du fichier cmak.cfg
$define=array("gcc"); # Toujours dclare (car il n'y a que gcc qui est support par cmak pour le moment... TODO)
if($^O=~/linux/i) {
	@define=(@define, "linux", "unix");
}
elsif($^O=~/darwin/i) {
	@define=(@define, "darwin","unix");
}
elsif($^O=~/freebsd/i) {
	@define=(@define, "freebsd","unix");
}
elsif($^O=~/win/i) {
	@define=(@define, "win32", "windows");
}
else {
	print STDERR "ATTENTION: Systeme d'exploitation non detecte. Pour le moment, cette version de cmak\n";
	print STDERR "           connait Linux, Darwin, FreeBSD et Windows.\n";
	print STDERR "           Le programme va supposer que vous avez un systeme UNIX.\n\n";
	@define=(@define,"unix");
}
*/

function find_dir_files ($dir)
{
	global $valid_extention;
		
	if(is_file($dir)) { cpp_file($dir); return;}

	$dir_array=scandir(addslashes($dir));
	
	foreach($dir_array as $key=>$value)
	{
		if($value != "." && $value != "..")
		{
			if(!is_dir($dir."/".$value)) cpp_file($dir."/".$value);
			else find_dir_files ($dir."/".$value);
		}
	}	
}


function arguments($argv) {
    $ARG = array();
    foreach ($argv as $arg) {
        if (strpos($arg, '--') === 0) {
            $compspec = explode('=', $arg);
            $key = str_replace('--', '', array_shift($compspec));
            $value = join('=', $compspec);
            $ARG[$key] = $value;
        } elseif (strpos($arg, '-') === 0) {
            $key = str_replace('-', '', $arg);
            if (!isset($ARG[$key])) $ARG[$key] = true;
        }
    }
    return $ARG;
} 

#
# Similaire  la commande print. Elle affiche le string lorsque
# le mode verbose est activ.
#
function vprint($bla) 
{
	if($verbose) {
		foreach ($bla as $key=>$value) { print $value; }
	}
}

#
# Mode interactif
# On va demander certaines questions  l'utilisateur
# avant de crer le Makefile (si l'argument -i existe)
#
function interactive_mode() 
{
	if($interactive!=0) {
		$makefile = cmak_prompt("Entrez le Makefile",$makefile);
		$executable = cmak_prompt("Entrez le nom du fichier executable",$executable);
	}
}

#
# Un petit prompt utilis par le mode interactif (-i).
#
# $1 = string question
# $2 = valeur par dfaut (quand l'utilisateur n'entre rien)
#
function cmak_prompt ($question,$default)
{
/*
	print $question. " ($default): ";

	# On va faire la lecture du clavier
	$v = <STDIN>;
	chomp($v); $v=&trim($v);
	if($v eq "") {
		$v=$default;
	}

	# On retourne enfin le rsultat
	return $v;
	*/
}

#
# Chercher le fichier C/C++ qui contient la fonction main(). 
# Si c'est le bon, le nom de l'excutable portera le nom de 
# ce fichier (sans l'extention bien sre).
#
function auto_detect_main ()
{
    global $cpp_file_list;
	
	# l'excutable doit tre vide
	if($executable != "") {
		return;
	}

	print("Recherche de la fonction \"main\"...\n");
	# On parcours la liste des fichiers C/C++
	foreach ($cpp_file_list as $key=>$value) {
		$v=$value;
		print("Scan du fichier \"$v\"...\n");
		if(detect_main($v)) 
		{
			print("Fonction trouvee dans le fichier \"$v\"...\n");
			// last; #c'est trouv!
		}
	}
}

#
# dtecter si main existe dans un fichier C/C++.
# La fonction retourne 1 si la fonction main() a t trouve.
#
# $1 = Fichier C/C++
#
function detect_main ($filename){
	global $executable;
	// my ($filename)=(@_);
	// my $line;

	# Si l'excutable est pleine... Pas la peine de faire
	# notre recherche
	if($executable != "") {
		return 0;
	}

	# On ouvre le fichier en lecture
	// echo 'BLAAA:\n'.$filename.'\n';
	$fp=fopen(addslashes($filename), "r") ; // || die("Erreur lors de l'ouverture du fichier $filename en lecture...\n");

	# On parcours ce mme fichier
	while(($line = fread($fp, 4096) ) ){		
		// chomp($line);
		# On teste si la ligne actuelle contient main
		# true si le rsultat est ok
		if(test_main($filename,$line)) {
			# Pour on ajoute le basename du filename (sans l'extention)
			$executable=basename(delext($filename));
			fclose($fp);
			return 1; #fin!
		}
	}
	fclose($fp);
	return 0;
}

#
# Permet de tester si une chane de caractres contient 
# une fonction main().
# 
# Si c'est le cas, elle retournera 1 et fera le ncessaire
# pour rendre le nom de l'excutable similaire au fichier 
# c/c++ contenant maint().
#
# $1 = Le nom du fichier fichier.cpp (sans le chemin)
# $2 = string (le contenu de la ligne lue depuis le fichier C/C++)
#
function test_main ($filename, $line_str)
{
	// my ($filename,$line_str) = (@_);
	# L'excutable doit tre vide
	// if($line_str=~/main/) {
	// if(strstr($line_str,"main")) {
	if(preg_match("/main([ ]*|)\(/", $line_str)) {
	    echo "\nFOUND main in ".$filename;
		return 1;
	}
	return 0;
}

#
# Dtecter automatiquement toutes les correspondances entre
# Headers<->CFLAGS|LDFLAGS en parcourant tous les fichiers
# C/C++.
#
function auto_detect_lib ($cpp_file_list)
{
	# On parcours la liste des fichiers C/C++
	foreach ($cpp_file_list as $key=>$value) {
		// my $v=$_;
		detect_lib($value,0);
	}
}

#
# Fonction appele par auto_detect_lib() pour trouver toutes
# les correspondances entre les *.h et CFLAGS/LDFLAGS.
# 
# Cette fonction peut parcourir rcursivement les headers.
#
# Elle ne scanne jamais un header deux fois (pour gagner du
# temps).
#
# $1 = fichier.h
# $2 = 0 si premier fichier. 1 Si deuxime (recursive). 2, etc.
# $3 = ignore error. 1 pour ignore les erreur. 0 sinon.
#
// my @visited_header; # Les headers qu'on as visit...
function detect_lib($filename,$recursive,$ignore_error)
{
global $visited_header;
global $cmak_cflags_list;
global $cmak_ldflags_list;
global $cmak_header_list;
	// my ($filename,$recursive,$ignore_error)=(@_);

	# Dans le cas ou l'option verbose a t dfinie
	if($verbose) {
		print "   ".$recursive;
		print "\"$filename\"\n";
	}

	# On ajoute ce lien dans visited
	foreach ($visited_header as $key=>$value) {
		if($value == $filename) {
			# dja visit...
			print("IGNORE: Le fichier $filename a dj t scann\n");
			return 1;
		}
	}
	array_push($visited_header,$filename);

	# On teste si l'on a atteint la fin du scan recursif...
	# 16 c'est suffisant!
	if($recursive>=16) {
		return 0; # MAXIMUM...
	}

	# Chercher le fichier .h selon son nom
	$filename = search_header($filename);
	
	# Dclaration du nom de fichier en forme de variable locale
	// my $HANDLE;

	# Ouverture du fichier pour le scan
	if(!$fp=fopen($filename,'r')) {
		print "Erreur lors de l'ouverture du fichier $filename en lecture...\nAjoutez un chemin de recherche pour les fichiers include avec l'option -I ou --include-dir\n";
		if(!$ignore_error) {
			exit (1);
		}
		return 1;
	}
	// my $line_str;
	while(($line_str=fread($fp,4096))) {
		// chomp($line_str);

		# S'il y a un include
		if(strstr($line_str,"#include",true)) {
			$include = trim($line_str);

			# On ne laisse que le fichier .h
			// a retoucher !!!!
			
			$include=preg_replace("/^\s*\#include\s*[\"<]/","", $include);
			$include=preg_replace("~s/[\">].*$//", "", $include);

			# et l on le teste avec tous les headers
			$i=0;
			foreach ($cmak_header_list as $key=>$value) {
				# Test
				if($include==$value) {
					# CFLAGS : il a t trouv dans la liste?
					if(defined($cmak_cflags_list[$i]) and $cmak_cflags_list[$i] != "") {
						$cflags.=" ".$cmak_cflags_list[$i];
						print("Le fichier $filename contient $include donc CFLAGS.=".$cmak_cflags_list[$i]."\n");
					}

					# LDFLAGS : il a t trouv dans la liste?
					if(defined($cmak_ldflags_list[$i]) and $cmak_ldflags_list[$i] != "") {
						$ldflags.=" ".$cmak_ldflags_list[$i];
						print("Le fichier $filename contient $include donc CFLAGS.=".$cmak_ldflags_list[$i]."\n");
					}

					# On vide le tout pour ne pas inclure deux fois la mme chose
					$cmak_cflags_list[$i]="";
					$cmak_ldflags_list[$i]="";

					$i=-1;
					// last; #FIN du test de tous les headers
				}

				# Incrmentation de $i afin d'avoir l'index  jour
				$i++;
			}

			# et enfin, on va faire une detect lib si celle-ci n'est 
			# pas dtecte ($i!=-1   ==   pas dtecte)
			if($i!=-1) {
				$ignore=0;
				# si <...> au lieu de "..." on ignore l'erreur
				# d'un include non trouv
				if($line_str=preg_match("/\#include.*<.*>/) { $ignore=1; }", $line_str))
				# On scanne le prochain fichier!
				detect_lib($include,$recursive+1,$ignore);
			}
		}
		
		# Si la ligne ne contient pas #include
		# on va faire un petit test:
		# Si le fichier c/c++ contient le mot main on va le choisir pour
		# qu'il soit le nom par dfaut du projet
		elseif($recursive==0) { // # rec=0 donc c le fichier pre. (le fich c/c++)
			if(test_main($filename,$line_str)) {
				# On ajoute le basename() du filename c/c++ (sans l'extention)
				$executable=basename(delext($filename));
			}
		}
	}
	fclose($fp);
}

#
# Chercher un header dans les chemins standard.
#
# $1 = fichier h
#
function search_header ($filename)
{
global $include_path_list;
	// my ($filename) = (@_);

	# On cherche dans -I
	foreach ($include_path_list as $key=>$value) {
		if(exists( $value."/".$filename)) {
			return $value."/".$filename;
		}
	}
	
	# On cherche dans include
	if(exists( "/usr/include/$filename")) {
		return "/usr/include/$filename";
	}

	# On cherche dans include linux specific
	if(cmak_defined("linux")) {
		if(exists( "/usr/include/linux/$filename")) {
			return "/usr/include/linux/$filename";
		}
	}

	# Dans le rpertoire local
	if(exists( "/usr/local/include/$filename")) {
		return "/usr/local/include/$filename";
	}

	# Dans le cas ou l'on a rien trouv
	# on retourne une valeur inchange
	return $filename;
}

#
# Chargement de tous les fichiers "cmak.cfg".
#
function load_cmak_cfg()
{
global $cmak_cfg;

	$loaded=0;

	# S'il y a un cfg personnalis qu'on doit charger
	if($cmak_cfg != "") {
		load_cmak_cfg_ex($cmak_cfg);
		$loaded=1;
	}
	
	# Dans le rpertoire actuel
	if(file_exists( "./cmak.cfg") ){
		load_cmak_cfg_ex("./cmak.cfg");
		$loaded=1;
	}

	# Cas d'un systme unix (Linux ou MacOS-X)
	if(cmak_defined("unix")) {
		# $HOME/.cmak.cfg
		/*
		$home = $_ENV("HOME");
		if($home != "") {
			if(file_exists( "$home/.cmak.cfg") ){
				load_cmak_cfg_ex("$home/cmak.cfg");
				$loaded=1;
			}
		}*/

		# et enfin le gnral...
		if(file_exists("/etc/cmak.cfg") ){
			load_cmak_cfg_ex("/etc/cmak.cfg");
			$loaded=1;
		}

		# et enfin le gnral...
		print "trying /usr/share/cmak/cmak.cfg\n";
		if(file_exists("/usr/share/cmak/cmak.cfg") ){
			load_cmak_cfg_ex("/usr/share/cmak/cmak.cfg");
			$loaded=1;
		 }

		# et enfin le gnral...
		print "trying usr:share/cmak/cmak.cfg\n";
		if(file_exists("usr:share/cmak/cmak.cfg") ){
			load_cmak_cfg_ex("usr:share/cmak/cmak.cfg");
			$loaded=1;
		}

	}

	# Cas d'un systme Windows
	elseif(cmak_defined("win32")) {
		if(file_exists( "c:\\cmak.cfg") ){
			load_cmak_cfg_ex("c:\\cmak.cfg");
			$loaded=1;
		}
	}

	# il n'a pas t trouv... :-(
	if($loaded==0) {
		print "Le fichier de configuration cmak.cfg est introuvable...\n";
		exit (1);
	}
}

#
# Chargement d'un fichier cfg.
# 
# $1 = chemin + nom du fichier cmak.cfg
#
function load_cmak_cfg_ex ($cmak_filename)
{
global $cmak_header_list;
global $cmak_ldflags_list;
global $cmak_cflags_list;
	// my ($cmak_filename)=(@_); # par dfaut c'est vide

	# Si le cmak filename a t trouv
	print "Chargement de la configuration du fichier $cmak_filename\n";

	# Variables
	$line = 0; # numro de ligne
	$ifdef_list = array(); # liste des ifdef (push, pop selon ifdef/endif)
	$ignore_cmd=0; # s'il y a un ifndef alors on ignore toutes les cmds header

	# Ouverture du fichier
	$fp=fopen(addslashes($cmak_filename), "r") || die("Erreur lors de l'ouverture du fichier $cmak_filename\n");
	$l; # La ligne lue
	$line_str; # Copie de $l utilise en bas dans la boucle
	while(($l=fread($fp,4096))) {
		// chomp($l);

		$line_str=$l; #sauvegarde de la ligne dans line_str

		# Quelques premires modifications
		$line++;
		$l=preg_replace("/\#.*$/","",$l);     # On enlve le commentaire le commentaire aprs une #
		$l=trim($l);  # On va enlever les espaces de gauche et de droite

		# Traitement de la commande + arguments
		$command = $l;
		$arg = $l;
		$command=preg_replace("/^([^\s]*).*$/$1/","",$command);
		$command=rtrim($command);
		$arg=preg_replace("/^[^\s]*/","", $arg);
		$arg=trim($arg);

		# cmd: vide (ON IGNORE!)
		if($command == "") { next; }

		# cmd: ifdef 
		if($command == "ifdef") {
			$ignore_cmd = (cmak_defined($arg))?0:1;
			array_push($ifdef_list,$arg);
		}

		# cmd: endif
		elseif($command == "endif") {
			$ifdef_count=count($ifdef_list);
			
			# on n'accepte pas d'arguments dans endif
			if($arg=preg_match("/[^\s].*/", $arg)) {
				cmak_error($cmak_filename,$line,$line_str,"endif n'accepte aucun argument.");
			}

			# S'il y a erreur (ifdef non dclar)
			if($ifdef_count<=0) {
				cmak_error($cmak_filename,$line,$line_str,"endif est declaree sans ifdef.");
			}

			# On enlve le prcdent lment de la liste
			array_pop($ifdef_list);

			$ignore_cmd=0; # par dfaut on ignore pas

			# Si il y a encore un define dans la liste on va la tester pour savoir si
			# on ignore ou pas
			if(count($ifdef_list)>0) {
				$ignore_cmd = (cmak_defined($ifdef_list[count($ifdef_list)-1]))?1:0;
			}
		}

		# Si on doit ignore les commandes c'est ici
		elseif($ignore_cmd==1) {
			next;
		}

		# define
		elseif($command == "define") {
			cmak_define($arg);
		}

		# enfin
		# cmd: header
		elseif($command == "header") {
			$table = decode (":",$arg);

			# nombre d'arguments
			if(count($table)>3 or count($table)<2) {
				cmak_error($cmak_filename,$line,$line_str,"La commande \"header\" n'accepte que 2 ou 3 arguments.");
			}
			
			# Lecture des arguments
			// my ($arg_header,$arg_ldflags,$arg_cflags) = (@table);
			$arg_header=$table[0];
			$arg_cflags=$table[1];
			$arg_ldflags=$table[2];
			$arg_header=trim($arg_header);
			$arg_cflags=trim($arg_cflags);
			$arg_ldflags=trim($arg_ldflags);
			
			# Ajout dans la liste
			array_push($cmak_header_list,$arg_header);
			array_push($cmak_ldflags_list,$arg_ldflags);
			array_push($cmak_cflags_list,$arg_cflags);
		}

		# Erreur + quitter
		else {
			cmak_error($cmak_filename,$line,$line_str,"Commande non reconnue");
		}
	}
	fclose($fp);
}
  
#
# Erreur dans le fichier cmd!
#
# $1 : fichier
# $2 : ligne
# $3 : contenu de la ligne
# $4 : explication (optionnel)
#
function cmak_error ($cmak_filename,$line,$line_str,$explication)
{
	// my ($cmak_filename,$line,$line_str,$explication) = (@_);
	print  "Erreur dans le fichier $cmak_filename\n";
	print  "Ligne:       $line\n";
	print  "Contenu:     ".trim($line_str)."\n";
	if($explication != "") {
		print  "Explication: ".$explication."\n";
	}
	exit (1);
}

#
# Ajouter une define cmak  la liste.
#
# $1 = define(str)
#
function cmak_define($bla)
{
	global $define;
	array_push($define,$bla);
}

#
# Pour savoir si une macro cmak a t dfinie.
#
# $1 = "constante"
#
function cmak_defined ($const)
{
	global $define;
	// my ($const) = (@_);
	foreach ($define as $key=>$value) {
		if($const == $value) {
			return 1;
		}
	}
	return 0;
}

#
# Crer un Makefile en utilisant toutes les variables
# modifies par les prcdentes fonctions.
#
function create_makefile()
{
	global $executable;
	global $cpp_source;
	global $cc;
	global $cflags;
	global $ldflags;
	global $cpp_file_list;
	global $makefile;
	global $default_path;
	
	
	# Dans le cas ou le fichier excutable n'est pas spcifi
	# par l'utilisateur et n'est pas dtect:
	# "On va mettre main par dfaut"
	if($executable == "") {
		$executable = "main";
	}
	
	# On change de compilateur si c'est important
	if($cpp_source) {
		// my @split=split(/:/,$cc);
		$split=explode(":",$cc);
		$cc=$split[1]; # Compilateur C++ (aprs les :)
	}
	else {
	    // my @split=split(/:/,$cc);
		$split=explode(":",$cc);		
		$cc=$split[0]; # Compilateur C
	}
	
	# Cration du makefile
	print "Creation du fichier \"$makefile\"...\n";

	# On ouvre le Makefile en criture
	$fp=fopen(addslashes($makefile),"w"); // || die("Erreur lors de l'ouverture du fichier $makefile en ecriture...\n");

	# Dbut du fichier Makefile
	fwrite($fp,"#--------------------------------------------------------------\n");
	fwrite($fp,"# Makefile generated with cmak version $VERSION.\n");

	# Insertion de l'heure et la date
	// my ($sec,$min,$hour,$mday,$mon,$year) = localtime(time);	

	fwrite($fp,"# Date: $mday/$mon/".($year+1900)." $hour:$min:$sec\n");

	# Fin du premier commentaire dans le Makefile
	fwrite($fp,"#--------------------------------------------------------------\n\n");

	# Les variables importantes du Makefile
	fwrite($fp,"PREFIX  = /usr/local\n");
	fwrite($fp,"CFLAGS  = $cflags\n");
	fwrite($fp,"LDFLAGS = $ldflags\n\n");

	fwrite($fp,"CC = $cc\n");
	fwrite($fp,"RM = rm -f\n");
	fwrite($fp,"INSTALL_PROG = install -m 755 -s\n\n");
	
	fwrite($fp,"EXE = $executable\n");

	# EXE NAME
	$exe_base=basename($executable);
	if($exe_base != $executable) {
		fwrite($fp,"EXE_BASE = ".$exe_base."\n\n");
	}
	else {
		fwrite($fp, "\n");
	}

	# On va mettre la liste des fichiers OBJ
	fwrite($fp,"OBJS =");
	foreach ($cpp_file_list as $key=>$value) {
		fwrite($fp," ");

		# Ajoute l'objdir
		if($objdir != "") {			
			fwrite($fp,$objdir."/");
		}

		# Ajoute le fichier o
		$value=str_replace($default_path, "",$value);
		if($value[0]=='/') $value = substr($value, 1, strlen($value));
		fwrite($fp,delext($value).".o");
	}
	
	# ALL
	fwrite($fp,"\n\nALL : \$(EXE)\n\n");

	# Ajoute toutes les compilations
	foreach ($cpp_file_list as $key=>$value) {
		#Cration du nom du fichier obj
		$value=str_replace($default_path, "",$value);
		if($value[0]=='/') $value = substr($value, 1, strlen($value));
		$obj=delext($value).".o";

		# Ajoute le chemin du fichier obj (si c'est dfini)
		if($objdir != "") {
			$obj=$objdir."/".$obj;
		}
		
		#criture dans le Makefile
		fwrite($fp,"$obj : $value\n");
		fwrite($fp,"\t\$(CC) -c $value \$(CFLAGS) -o $obj\n\n");
	}

	# Ajoute maintenant la linkage
	fwrite($fp,"\$(EXE) : \$(OBJS)\n");
	fwrite($fp,"\t\$(CC) \$(OBJS) -o \$(EXE) \$(LDFLAGS)\n\n");

	# install: (systme unix seulement)
	if(cmak_defined("unix")) {
		fwrite($fp,"install : \$(EXE)\n");
		fwrite($fp,"\t\$(INSTALL_PROG) \$(EXE) \$(PREFIX)/bin\n\n");
	}

	# uninstall:
	fwrite($fp,"uninstall :\n");
	if($executable != $exe_base) {
		fwrite($fp,"\t\$(RM) \$(PREFIX)/bin/\$(EXE_BASE)\n\n");
	}
	else {
		fwrite($fp,"\t\$(RM) \$(PREFIX)/bin/\$(EXE)\n\n");
	}

	# clean:
	fwrite($fp,"clean :\n");
	fwrite($fp,"\t\$(RM) \$(OBJS) \$(EXE)");

	# Fermeture
	fclose($fp);
}

#
# Lecture de tous les arguments.
#
# Cette fonction quitte s'il y a une erreur.
#
function get_arguments()
{
   global $cpp_file_list;
   global $ARGV;
   global $interactive;
   global $verbose;
   global $detect_lib;
   global $VERSION;
   global $objdir;
   global $cmak_cfg;
   global $cflags;
   global $ldflags;
   global $include_path_list;
   global $executable;
   global $default_path;
   
	$fin_option=0; # si -- alors aucun option ne sera lue
	
	# Si l'utilisateur ne dfinit aucune option on va lui afficher l'aide
	if(count($ARGV)==0) {
		help();
	}
	
	# On parcours toutes les options
	for($i=0;$i<count($ARGV);$i++) {
		$bla=$ARGV[$i];
		# NB: for($i) au lieu de foreach() car on va utiliser ++$i pour avoir le prochaine arg
		
		# Pour en finir avec les options
		if($bla == "--") {
			$fin_option=1;
			next;
		}
		
		# Si les options sont finies alors on ajoute les fichier C/CPP seulement
		if($fin_option) {
			cpp_file($bla);
			next;
		}
		
		#* Ici on test les options mono
		
		# --detect-lib (dtection automatique de librairie)
		elseif($bla == "--detect-lib" or $bla == "-dl") {
			$detect_lib=1;
		}

		# --version
		elseif($bla == "--version") {
			print $VERSION."\n";
			exit;
		}
		# --interactive
		elseif($bla == "--interactive" or $bla == "-i") {
			$interactive=1;
		}

		# --verbose
		elseif($bla == "--verbose" or $bla == "-v") {
			$verbose=1;
		}

		# Aide --help ou -h
		elseif($bla == "--help" or $bla == "-h") {
			help();
		}

		#* On teste les autres options. ex: -e value

		# --obj-dir
		elseif($bla == "-od" or $bla == "--obj-dir") {
			$objdir=$ARGV[++$i];
		}

		# --cflags
		elseif($bla == "-C" or $bla == "--cflags") {
			$cflags.=" ".$ARGV[++$i];
		}

		# --ldflags
		elseif($bla == "-LD" or $bla == "--ldflags") {
			$ldflags.=" ".$ARGV[++$i];
		}

		# --lib-dir
		elseif($bla == "-L" or $bla == "--lib-dir") {
			$cflags.=" -L".$ARGV[++$i];
		}

		# --include-dir
		elseif($bla == "-I" or $bla == "--include-dir") {
			$cflags.=" -I".$ARGV[++$i];
			array_push($include_path_list,$ARGV[$i]);
		}


		# choisir le chemin du fichier config
		elseif($bla == "--cfg") {
			$cmak_cfg=$ARGV[++$i];
		}
		
		# --executable
		elseif($bla == "-e" or $bla == "--executable") {
			$executable=$ARGV[++$i];
		}

		# --makefile
		elseif($bla == "-m" or $bla == "--makefile") {
			$makefile=$ARGV[++$i];
		}

		# Si ce n'est pas une option alors ajouter le fichier .cpp
		// Voir pour grer le cas *.c..
		else {
			// cpp_file($bla);
			$default_path=$bla;
			find_dir_files ($bla);
			next;
		}

		# si c'est une option alors tester s'il y a eu un dpassement
		if($i>=count($ARGV)) {
			print "Erreur dans l'argument ".($i)." contenant \"".$ARGV[$i-1]."\"...\n";
			print "--help pour plus d'informations.\n";
			exit (1);
		}
	}

	# on doit absolument entrer un fichier c/c++
	if(count($cpp_file_list) == 0) {
		print "Vous devez au moins entrer un fichiers C/C++ dans les arguments.\n";
		print "--help pour plus d'informations\n";
		exit (1);
	}
}

#
# Enlve l'extention d'un nom de fichier.
# Le rsultat est retourn.
#
# $1 = fichier avec extention
#
function delext($filename)
{
	// my ($filename) = (@_);
	if(!empty($filename)) {
		echo $filename=preg_replace("/\.[^.]*$/" , "", $filename);
	}
	return $filename;
}

#
# Ajoute un fichier C/C++  la liste.
# Quitte s'il y a une erreur.
#
# $1 = fichier c/c++
#
function cpp_file ($filename)
{
	global $cpp_file_list;
	global $valid_extention;
	// my ($filename) = (@_);
	//my $ext;

	# On teste l'existence du fichier
	if(!file_exists($filename)) {
		print "Le fichier \"$filename\" n'existe pas...\n";
		exit (1);
	}

	# On prend l'extention
	$ext=$filename;
	$ext=preg_replace("/^.*\./", "",$ext);

	// echo "EXTENSION :".$ext;
	
	# On teste le fichier (si l'extention est bonne)
	$invalid_extention=1;
	if(preg_match("/\./",$filename)) {
		foreach ($valid_extention as $key=>$value){
			if($value == $ext) {
				
				$invalid_extention=0;
				# Test si c ou c++
				$cpp_source=($ext == "c")?0:1;
				last;
			}
		}
	}
	
	# Si l'extention n'est pas bonne
	if($invalid_extention) {
		print "Le fichier \"$filename\" ne contient pas une extention c/c++ valide.\n";
		print "Les extentions acceptes: ".implode(", ",$valid_extention)."\n";
		// exit (1);
	}else
	{
	# on ajoute le fichier  la liste
		array_push($cpp_file_list,$filename);
	}
	
}

#
# Une petite aide + exit
#
function help()
{
	global $VERSION;
	# Petite aide pour l'utilisateur
	print "cmak version $VERSION, Generateur automatique de Makefile\n";
	print "Auteur: Achraf cherti <achrafcherti\@gmail.com>\n";
	print "\n";
	print "Les accents ont ete enleves pour eliminer les conflits dans l'encodage.\n";
	print "(par exemple entre: UTF-8 et ISO-8859-1).\n";
	print "\n";
	print "Syntaxe: cmak [options] fichier1.cpp fichier2.c etc...\n";
	print "\n";
	print "Options:\n";
	print "    --version                       Affiche la version du logiciel\n";
	print "    -h,  --help                     Aide sur les options\n";
	print "    -v,  --verbose                  Afficher plus d'informations que la normale\n";
	print "    -i,  --interactive              Mode interactif (demande des questions)\n\n";

	print "    -dl, --detect-lib               Detecte automatiquement les bibliotheques\n";
	print "                                    selon les headers dans le code source.\n";
	print "                                    Editez le fichier cmak.cfg pour\n";
	print "                                    personnaliser cette detection\n";
	print "\n";
	print "    --cfg <chemin/fichier.cfg>      Choisir manuellement le chemin de cmak.cfg\n";
	print "\n";
	print "    -m,  --makefile   <fichier>     Le fichier Makefile (par defaut 'Makefile')\n";
	print "\n";
	print "    -e,  --executable <fichier>     Le fichier executable genere apres la\n";
	print "                                    compilation\n";
	print "\n";
	print "    -od, --obj-dir <repertoire>     Le repertoire ou seront mis les fichiers .o\n";
	print "\n";
	print "    -L,  --lib-dir <repertoire>     Cette option peut etre definie plusieurs\n";
	print "                                    fois. Elle permet d'indiquer le chemin\n";
	print "                                    des bibliotheques .a\n";
	print "\n";
	print "    -I,  --include-dir <rep>        Cette option peut etre definie plusieurs\n";
	print "                                    fois. Elle permet d'indiquer le chemin des\n";
	print "                                    fichier .h\n";
	print "\n";
	print "    -C,  --cflags <flags>           Ajouter aux CFLAGS une option (peut etre\n";
	print "                                    definie plusieurs fois)\n\n";
	
	print "    -LD, --ldflags <flags>          Ajouter aux LDFLAGS une option (peut etre\n";
	print "                                    definie plusieurs fois)\n";
	print "\n";

	# Et enfin, on quitte ;-)
	exit (0);
}

#
# enlever les espaces  gauche et  droite d'un string
#
/*
sub trim
{
	my ($string) = (@_);
	return &ltrim(&rtrim($string));
}
*/
#
# Enlever les espaces  gauche d'un string
#
/*
sub ltrim
{
	my ($string) = (@_);
	if(defined($string)) {
		$string =~ s/^\s+//;
	}
	return $string;
}
*/
#
# Enlever les espaces  droite d'un string
#
/*
sub rtrim
{
	my ($string) = (@_);
	if(defined($string)) {
		$string =~ s/\s+$//;
	}
	return $string;
}
*/

# On prend les arguments @ARGV
// $ARGV = arguments($_SERVER["argv"]);
$ARGV = array_slice($_SERVER["argv"],1);
var_dump($ARGV);

get_arguments();

# Dois-on dtecter automatiquement les bibliothques de sous programmes ?
if($detect_lib) {
	# On charge le(s) fichier(s) cmak.cfg
	load_cmak_cfg();

	# On va alimenter les CFLAGS et les LDFLAGS
	# en utilisant les fichiers *.h inclus dans tout le code source
	auto_detect_lib();

	# Et enfin, on va donner un petit rsultat affich aprs la dtection 
	# automatique des CFLAGS et LDFLAGS 
	$cflags  = trim($cflags);
	$ldflags = trim($ldflags);

	print "CFLAGS  = $cflags\n";
	print "LDFLAGS = $ldflags\n\n";
}
else {
	# On cherche dans tous les fichiers C/C++ ou se trouve la fonction main()
	auto_detect_main();

	$cflags  = trim($cflags);
	$ldflags = trim($ldflags);
}

# Dans le cas ou l'utilisateur a demand le mode interactif
# on va lui poser certaines questions...
interactive_mode();

# On va enfin gnrer le Makefile
create_makefile();


?>
